import createPub from '../';
import {usePub}  from '../';

export type QueryFuncType = (...args: Array<any>) => Promise;

export type CreateQueryPubOptions = {
    queryFunc?: QueryFuncType,
    cacheTimeout?: number,
    hidePendingOnRefresh?: boolean
    avoidRepeatInProgress?: boolean,
    forceRefreshAfterTimeout?: boolean,
}

enum PromiseStatus  {
    PENDING = "PENDING",
    FULFILLED = "FULFILLED",
    REJECTED = "REJECTED"
}

type PromiseSnapshot = {
    status: PromiseStatus,
    payload: any,
    error: Error
}

const defaultOptions:CreateQueryPubOptions = {
    cacheTimeout: 2 * 60 * 1000,
    hidePendingOnRefresh: true,
    avoidRepeatInProgress: true,
    forceRefreshAfterTimeout: true,
}

export default (globalOptions:CreateQueryPubOptions) => {
    const queryPub        = createPub({queries: {}, mutations: {}})
    const currentPromises =           {queries: {}, mutations: {}}

    globalOptions.cacheTimeout              ||= defaultOptions.cacheTimeout

    globalOptions.hidePendingOnRefresh        = globalOptions.hidePendingOnRefresh     === undefined ? defaultOptions.hidePendingOnRefresh     : globalOptions.hidePendingOnRefresh
    globalOptions.avoidRepeatInProgress       = globalOptions.avoidRepeatInProgress    === undefined ? defaultOptions.avoidRepeatInProgress    : globalOptions.avoidRepeatInProgress
    globalOptions.forceRefreshAfterTimeout    = globalOptions.forceRefreshAfterTimeout === undefined ? defaultOptions.forceRefreshAfterTimeout : globalOptions.forceRefreshAfterTimeout


    /*
     * function to remember basic transport function (like fetch or gql) in closure for further usage
     *
     * @param {function} queryFunc? - basic transport function. If omitted, used function from options of createQueryPub
     * @returns {function} "createQuery", which receives promise name for pub, isMutation flag, cacheTagsFunc and query function arguments 
     */
    const withQueryFunc = (queryFunc?: QueryFuncType):Function => {
        //TODO: type params
        return ({promiseName, name, isMutation, cacheTagsFunc, args=[], cacheTimeout, hidePendingOnRefresh, avoidRepeatInProgress, forceRefreshAfterTimeout}) => {
            promiseName = promiseName || name

            const qF = queryFunc || options.queryFunc
            const defaultArgs = args

            cacheTimeout ||= globalOptions.cacheTimeout

            hidePendingOnRefresh        = hidePendingOnRefresh     === undefined ? globalOptions.hidePendingOnRefresh     : hidePendingOnRefresh
            avoidRepeatInProgress       = avoidRepeatInProgress    === undefined ? globalOptions.avoidRepeatInProgress    : avoidRepeatInProgress
            forceRefreshAfterTimeout    = forceRefreshAfterTimeout === undefined ? globalOptions.forceRefreshAfterTimeout : forceRefreshAfterTimeout

            /*
             * Concats default arguments from withQueryFunc resulting function with passed arguments
             * 
             */
            const mixArgs = (args:any[]) => 
                Array.isArray(defaultArgs) ? 
                    [...defaultArgs, ...args] :
                    (typeof defaultArgs === 'function' ? 
                        defaultArgs(...args) : args)

            /*
             * used to create key for pub
             */
            const stringifyArgs = args => 
                JSON.stringify(mixArgs(args))

            /*
             * invokes query, using basic transport function and mixArgs
             */
            const makeQuery = (...args:any[]) => {
                return qF(...mixArgs(args))
            }

            /*
             * Checks tags intersection for invalidation of some cached query result
             * TODO: type it 
             */
            const intersectTags = (first:Array<any>|any, second:Array<any>|any) => {
                first  = Array.isArray(first)  ? first  : [first]
                second = Array.isArray(second) ? second : [second]

                return first.find(firstItem => {
                    if (!firstItem.id || !firstItem.type) throw new Error('cacheTagsFunc doesn\'t returns object with id and type properties: ' + JSON.stringify(first, null, 4))
                    return second.find(secondItem => { 
                        if (!secondItem.id || !secondItem.type) throw new Error('cacheTagsFunc doesn\'t returns object with id and type properties: ' + JSON.stringify(second, null, 4))
                        return firstItem.type === secondItem.type && firstItem.id === secondItem.id 
                    })
                })
            }


            const invalidateTags = tags => {
                if (!tags) return

                for (const promiseBranch of Object.values(queryPub.queries)){
                    for (const [args, promiseState] of Object.entries(promiseBranch))
                        if (promiseState.cacheTags && intersectTags(tags, promiseState.cacheTags)){
                            //console.log('invalidate', tags, args, promiseState)
                            promiseState.invalidated = true;
                            delete promiseBranch[args]
                        }
                }
            }


            /*
             * Performs query, if needed, using makeQuery or forceQuery
             * Checks is query mutation or pure query, and options like avoidRepeatInProgress or 
             * cache timeouts. So, in some circumstances, query skipped.
             */
            const query = (...args:any[]) => {
                if (isMutation){
                    if (avoidRepeatInProgress && queryPub.mutations[promiseName]?.refreshing){
                        return currentPromises.mutations[promiseName]
                    }
                    else {
                        if (!hidePendingOnRefresh || !queryPub.mutations[promiseName]){
                            queryPub.mutations[promiseName] = {status: 'PENDING', refreshing: true}
                        }
                        if (!queryPub.mutations[promiseName].refreshing){
                            queryPub.mutations[promiseName].refreshing = true
                        }
                        return currentPromises.mutations[promiseName] = makeQuery(...args).then(payload => queryPub.mutations[promiseName] = (invalidateTags(cacheTagsFunc && cacheTagsFunc(payload, args)),{status: 'FULFILLED', payload}),
                                                                                                error   => queryPub.mutations[promiseName] = {status: 'REJECTED', error})
                    }
                }
                if (!isMutation){
                    const stringifiedArgs = stringifyArgs(args)
                    if (((!cacheTimeout && queryPub.queries[promiseName]?.[stringifiedArgs].status === 'FULFILLED') || 
                        (queryPub.queries[promiseName]?.[stringifiedArgs] && (Date.now() - queryPub.queries[promiseName][stringifiedArgs].timeStamp) < cacheTimeout)) &&
                         !queryPub.queries[promiseName]?.[stringifiedArgs].invalidated){

                        return currentPromises.queries[promiseName][stringifiedArgs] 
                    }
                    return forceQuery(...args)
                }
            }

            /*
             * Forces query using makeQuery, but maintain pub structure. (cache timeout, status and so)
             * Always runs query for Mutation
             * The only reason to skip query for Query is avoidRepeatInProgress (deduplicate queries). Maybe this logic should be moved to query function
             *
             */
            const forceQuery = (...args:any[]) =>  {
                if (isMutation) return query(...args)



                const stringifiedArgs = stringifyArgs(args)

                if (!queryPub.queries[promiseName]) {
                    queryPub.queries[promiseName] = {}
                }
                if (!currentPromises.queries[promiseName]){
                    currentPromises.queries[promiseName] = {}
                }


                if (avoidRepeatInProgress && queryPub.queries[promiseName]?.[stringifiedArgs]?.refreshing){
                    return currentPromises.queries[promiseName][stringifiedArgs]
                }

                if (!hidePendingOnRefresh || !queryPub.queries[promiseName]?.[stringifiedArgs]){
                    queryPub.queries[promiseName][stringifiedArgs] = {status: 'PENDING', timeStamp: Date.now(), refreshing: true}
                }

                if (!queryPub.queries[promiseName][stringifiedArgs].refreshing){
                    queryPub.queries[promiseName][stringifiedArgs].refreshing = true
                }

                if (queryPub.queries[promiseName][stringifiedArgs].timeoutId){
                    clearTimeout(queryPub.queries[promiseName][stringifiedArgs].timeoutId)
                }



                return currentPromises.queries[promiseName][stringifiedArgs] = makeQuery(...args)
                                                                                    .then(payload => queryPub.queries[promiseName][stringifiedArgs] = {    
                                                                                                                                                            status: 'FULFILLED', 
                                                                                                                                                            payload, 
                                                                                                                                                            timeStamp: Date.now(), 
                                                                                                                                                            cacheTags: cacheTagsFunc && cacheTagsFunc(payload),
                                                                                                                                                            timeoutId: cacheTimeout && forceRefreshAfterTimeout ? setTimeout(() =>forceQuery(...args), cacheTimeout) : null
                                                                                                                                                        },
                                                                                           error   => queryPub.queries[promiseName][stringifiedArgs] = {
                                                                                                                                                            status: 'REJECTED', 
                                                                                                                                                            error, 
                                                                                                                                                            timeStamp: Date.now(),
                                                                                                                                                            timeoutId: cacheTimeout && forceRefreshAfterTimeout ? setTimeout(() =>forceQuery(...args), cacheTimeout) : null
                                                                                                                                                        })
            }

            /*
             * Creates hook name from promise name like rtk query
             */
            const hookName = `use${promiseName[0].toUpperCase() + promiseName.slice(1) + (isMutation ? 'Mutation' : 'Query')}`

            /*
             * Mutation hook. Returns query function and current pub data
             */
            const useMutation = () => {
                usePub(queryPub.mutations)

                return [query, queryPub.mutations[promiseName] || {status: 'INIT'}]
            }

            /*
             * Query hook. Calls query function with arguments passed. 
             * Listening for pub.
             * @returns pub data
             */
            const useQuery = (...args) => {
                query(...args)

                usePub(queryPub.queries[promiseName])

                const stringifiedArgs = stringifyArgs(args)
                return queryPub.queries[promiseName][stringifiedArgs]
            }

            /*
             * Gets sub pub for this promiseName and this parameters. Used by NamedRoute
             * to subscribe on pub and refresh when pub updates. Working only for Query,
             * not for Mutations
             */
            query.getQueryPubBranch = (...args) => {
                const stringifiedArgs = stringifyArgs(args)
                return [queryPub.queries[promiseName], stringifiedArgs] 
            }

            return {makeQuery, query, forceQuery, [hookName]: isMutation ? useMutation : useQuery, hook: isMutation ? useMutation : useQuery, hookName}
        }
    }


    const flushCache = () => {
        for (const [promiseName,promiseBranch] of Object.entries(queryPub.queries)){
            for (const [args, promiseState]    of Object.entries(promiseBranch)){
                if (promiseState.timeoutId){
                    clearTimeout(promiseState.timeoutId)
                }
                promiseState.invalidated = true
                delete promiseBranch[args]
                //promiseState.invalidated = true
            }
            //queryPub[promiseName] = {}
        }

        //console.log(queryPub.mutations)
        for (const [promiseName,promiseBranch] of Object.entries(queryPub.mutations)){
            //delete queryPub.mutations[promiseName]
            queryPub.mutations[promiseName] = {status: 'INIT'}
            //promiseBranch.invalidated = true
        }
        //console.log('AFTER flush',JSON.stringify(queryPub, null, 4))
    }

    return {queryPub, withQueryFunc, ...(globalOptions.queryFunc ? {createQuery: withQueryFunc(globalOptions.queryFunc)}: {}), flushCache}
}
