import { ComponentType, Fragment } from 'react';
import { GetUserRolesFunc, NamedRoute, NamedLink, createPrivateRoute, addRoute } from '../router';
import { Pub } from '../pub';
import { createQueryPub } from '../pub';

export type QueryRoute = {
    //query part
    args?: any[]|Function,
    cacheTagsFunc?: Function,
    //if set, routing can't work
    isMutation?: boolean,
    /*
    //it aren't works for now
    query: QueryFuncType,
    */
    cacheTimeout?: number,
    hidePendingOnRefresh?: boolean
    avoidRepeatInProgress?: boolean,
    forceRefreshAfterTimeout?: boolean,



    //route part
    path? : string,
    component? : ComponentType,

    //if defined, Route component created for this route are private
    roles?: string[],
    fallback?: string|Function


    //pending, error and prop names configuration override on route level
    Pending?: ComponentType,
    Error?:   ComponentType,

    errorQueryComponentErrorProp?: string,
    componentQueryResultProp? : string,
}

export type QueryRoutes = {
    [name: string]: QueryRoute
}

export type CreateQueryRoutesOptions = {
    //global query configuration, basic query function are mandatory
    query: QueryFuncType,
    cacheTimeout?: number,
    hidePendingOnRefresh?: boolean
    avoidRepeatInProgress?: boolean,
    forceRefreshAfterTimeout?: boolean,

    //global route configuration, pending, error and prop names configuration
    Pending?: ComponentType,
    Error?:   ComponentType,
    errorQueryComponentErrorProp?: string,
    componentQueryResultProp? : string,

    //global private routing configuration
    getUserRoles: GetUserRolesFunc, //function to get roles of current user
    fallback: string|Function //function/string to be redirected to in case of lack of permission
    routeRoles: string[] //default roles (like ['anon']

    queryRoutes: QueryRoutes
}

export type QueryRouteResult = {
    //results about routing:
    
    //NamedRoute decorated
    Route: ComponentType,
    //Same value as for Route, but named.  For example, if name "category" it will be CategoryRoute 
    [name: string]: ComponentType, 
    //NamedLink to exactly this route
    Link: ComponentType,
    //Same value as for Link, but named. Like CategoryLink
    [name: string]: ComponentType, 


    //results about queries
    makeQuery: Function,
    query: Function,
    forceQuery: Function,
    flushCache: Function,
    hook: Function,

    //named hook useNameQuery
    [name: string]: ComponentType, 
}

export type QueryRoutesResults = {
    [name: string]: QueryRouteResult
}

export type CreateQueryRoutesResult = {
    //Global results:
    //routing:
    NamedRoute: ComponentType,
    PrivateRoute: ComponentType,
    NamedLink   ; ComponentType, //just copy of NamedLink, nothing special
    //Component renders all routes in one place. Shortcut, but isn't flexible if you want routes
    //in different places of page layout
    AllRoutes: ComponentType,

    //queryPub
    queryPub: Pub,
    withQueryFunc: Function, //type it
    createQuery: Function, //type it
    
    queryRoutes: QueryRoutesResults
}

export const createQueryRoutes = (options:CreateQueryRoutesOptions):CreateQueryRoutesResult => {
    //global initialization:
    //ROUTING
    const { Pending, Error:ErrorComponent, errorQueryComponentErrorProp:errorProp, componentQueryResultProp:resultProp} = options

    const Route:ComponentType = ({pendingQueryComponent, errorQueryComponent, errorQueryComponentErrorProp, componentQueryResultProp, ...props}) => {
        pendingQueryComponent ||=  Pending
        errorQueryComponent   ||=  ErrorComponent
        errorQueryComponentErrorProp ||= errorProp
        componentQueryResultProp ||= resultProp

        return <NamedRoute {...props} 
                  pendingQueryComponent={pendingQueryComponent} 
                  errorQueryComponent={errorQueryComponent} 
                  errorQueryComponentErrorProp={errorQueryComponentErrorProp} 
                  componentQueryResultProp={componentQueryResultProp}/>
    }

    const {getUserRoles, fallback, routeRoles} = options

    const PrivateRoute:ComponentType = createPrivateRoute(getUserRoles, fallback, routeRoles)(Route)

    //queryPub 
    const { query, cacheTimeout, hidePendingOnRefresh, avoidRepeatInProgress, forceRefreshAfterTimeout } = options 
    const queryPubResult = createQueryPub({ queryFunc: query, cacheTimeout, hidePendingOnRefresh, avoidRepeatInProgress, forceRefreshAfterTimeout })
    

    const queryRoutesResults:QueryRoutesResults = {};

    for (const [name, queryRoute] of Object.entries(options.queryRoutes)){
        const {path, args, component, Pending, Error:ErrorComponent, errorQueryComponentErrorProp:errorProp, componentQueryResultProp:resultProp, roles, fallback,
                cacheTagsFunc,  isMutation, cacheTimeout, avoidRepeatInProgress, hidePendingOnRefresh, forceRefreshAfterTimeout} = queryRoute

        const camelCaseName = name[0].toUpperCase() + name.slice(1)

        addRoute({name, path});

        let queryRouteResult:QueryRouteResult = {}

        if (args){
            const createQueryResult = queryPubResult.createQuery({name, isMutation, cacheTagsFunc, args, cacheTimeout, avoidRepeatInProgress, hidePendingOnRefresh, forceRefreshAfterTimeout })
            const {makeQuery, query, forceQuery, hook, hookName} = createQueryResult

            queryRouteResult = {
                makeQuery, query, forceQuery, hook, [hookName]: hook
            }
        }

        if (path && component && isMutation){
            throw new Error(`You can't use mutation for route "${name}"`)
        }

        if (path && component && !isMutation){
            const R:ComponentType = ({pendingQueryComponent, errorQueryComponent, errorQueryComponentErrorProp, componentQueryResultProp, name:routeName, path: p, component: c, roles: r, fallback: f, query: q,...props}) => {
                pendingQueryComponent           ||=  Pending
                errorQueryComponent             ||=  ErrorComponent
                errorQueryComponentErrorProp    ||=  errorProp
                componentQueryResultProp        ||=  resultProp
                routeName                       ||=  name
                p                               ||=  path
                c                               ||=  component
                r                               ||=  roles
                f                               ||=  fallback
                q                               ||=  queryRouteResult.query

                const RouteComponent = roles ? PrivateRoute : Route

                return <RouteComponent 
                            {...props} 
                            name={routeName}
                            path={p}
                            component={c}
                            pendingQueryComponent={pendingQueryComponent} 
                            errorQueryComponent={errorQueryComponent} 
                            errorQueryComponentErrorProp={errorQueryComponentErrorProp} 
                            componentQueryResultProp={componentQueryResultProp}
                            roles={r}
                            fallback={f}
                            query={q}
                            />
            }

            const L:ComponentType = ({name:linkName, ...props}) => {
                linkName ||= name
                return <NamedLink {...props} name={linkName}/>
            }


            queryRouteResult = {
                ...queryRouteResult,
                Route: R,
                [camelCaseName + 'Route']: R, 

                Link: L,
                [camelCaseName + 'Link']: L, 
            }
        }


        queryRoutesResults[name] = queryRouteResult
    }

    const AllRoutes = () => 
        (
            <>
                {Object.entries(queryRoutesResults).map(([key,{Route}]) => Route ? <Route key={key}/> : null)}
            </>
        )

    const {queryPub, withQueryFunc, createQuery, flushCache} = queryPubResult

    return {
        NamedRoute: Route,
        PrivateRoute,
        NamedLink,
        AllRoutes,

        queryPub,
        withQueryFunc,
        createQuery,
        flushCache,


        queryRoutes: queryRoutesResults,
    }
}
